/*
 * Copyright 2022-2024 the Pacemaker project contributors
 *
 * The version control history for this file may have further details.
 *
 * This source code is licensed under the GNU Lesser General Public License
 * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
 */

#ifndef PCMK__CRM_COMMON_SCORES_INTERNAL__H
#define PCMK__CRM_COMMON_SCORES_INTERNAL__H

int pcmk__add_scores(int score1, int score2);

#endif // PCMK__CRM_COMMON_SCORES_INTERNAL__H
